#pragma once

#include "../SMemory/SBlocks.h"
#include <assert.h>

//
//	class cPGPPacket
//		Accumulates data at temporary body.
//		Writes data to output packet partially at Final.
//
class cPGPPacket
{
	cPlainTextBlock m_body;
	BYTE m_tag;

public:
	cPGPPacket( BYTE tag = 0, BYTE *body = NULL, int len = 0 )
	{
		Init( tag );
		Add( body, len );
	}

	bool Init( BYTE tag )
	{
		m_tag = tag;
		m_body.Free();
		return true;
	}

	template < class T > bool Add( T t )
	{
		return Add( (BYTE*)&t, sizeof(T) );
	}

	bool Add( BYTE *body, int len )
	{
		return m_body.WriteBytes( body, len ) != 0;
	}

	BYTE *GetSpace( int len )
	{
		int pos = m_body.bufferSize;

		if ( m_body.WriteBytes( NULL, len ) == 0 )
			return NULL;

		assert( m_body.bufferSize == pos + len );

		return m_body.buffer + pos;
	}

	enum { maxStep = 0x200, maxStepGnuPG = 0x1000 }; // RFC 2440. The first partial length MUST be at least 512 octets long
	
	
	int getPartialSize( int size )
	{
		int i = 0;

		for( int s = size;  s > 1 ; i++ )
		{
			s /= 2;
		}

		int r = 1 << i; 

		//
		//	19.10.2015
		//	Keopatra hangs at size more ((
		//
		if ( r > 0x2000 )
			return 0x2000;

		return r;
	}

	//
	//	Final
	//	Make formatted packet to ptbout.
	//	The undefinedSize is required to satisfy K;eopatra software based on GnuPG
	//	It strange but it does not recognize decrypted message if compressed packet has new format size.
	//	It reports success if this packed has old format size type 3 ( undefined )
	//
	bool Final( cPlainTextBlock *ptbOut, bool undefinedSize = false )
	{
		int res;

		BYTE tag = m_tag;

		int step = maxStepGnuPG; // maxStep;
		int pos = 0;

		do
		{
			bool partialSize = false;

			if ( ( pos + step > m_body.bufferSize ) || undefinedSize )
				step = m_body.bufferSize - pos;
			else
			{
				step = getPartialSize( m_body.bufferSize - pos );
				partialSize = true;
			}

			int pktSize = 0;

			res = pgp_PacketHeader_Format( NULL, &pktSize, tag, m_body.buffer + pos, step, partialSize, undefinedSize );
			if ( res != ppe_bufferTooSmall )
				return false;
			
			int ptbOutPos = ptbOut->bufferSize;

			if ( ptbOut->WriteBytes( NULL, pktSize ) == 0 )
				return false;

			res = pgp_PacketHeader_Format( ptbOut->buffer + ptbOutPos, &pktSize, tag, m_body.buffer + pos, step, partialSize, undefinedSize );
			if ( res != ppe_ok )
				return false;

			// Next packet is partial type ( without 'Packet Tag' byte ) and cannot be undefined size.
			tag = 0;
			undefinedSize = false;

			pos += step;
		}
		while( pos < m_body.bufferSize );

		return true;
	}
};
