#include "cAlgRC2.h"
#include "RC2/rc2.h"

#define BLOCK_SIZE 8
#define KEY_LENGTH 40 // bits size
#define KEY_LENGTH_MAX 128 * 8// bits size

cAlgRC2::cAlgRC2( int keyBitLength )
{
	m_keySize = keyBitLength / 8;
	m_extKey = NULL;
}

cAlgRC2::~cAlgRC2()
{
	if ( m_extKey != NULL )
		safeFree( m_extKey );
}

BOOL cAlgRC2::getAlgConstants( int *blockSize, int *keySize )
{
	*blockSize = BLOCK_SIZE;
	*keySize   = m_keySize;
	return TRUE;
}


int  cAlgRC2::getBlockSize()
{
	return BLOCK_SIZE;
}

bool cAlgRC2::expandKey(
		BYTE *key,
		int keySizeBytes, // key size in bytes
		int effectiveKeyLengthBits // effective key length in bits = 1024
	)
{
	if ( m_extKey != NULL )
		safeFree( m_extKey );

	if ( keySizeBytes != m_keySize )
		return false;

	m_extKey = safeAlloc( sizeof(RC2_Schedule) );

	if (!m_extKey)
		return false;

	rc2_keyschedule( (RC2_Schedule *)m_extKey, key, keySizeBytes, effectiveKeyLengthBits );

	return true;
}

int cAlgRC2::encryptECB(
		const BYTE *inBf,	// Input plaintext buffer. 
		BYTE *outBf,		// Output encrypted buffer
		int  bfSize			// In/out buffers' size.
		)
{
	if ((bfSize % BLOCK_SIZE) != 0) return sae_invalidBlockSize;

	for( int i = 0; i < bfSize; i += BLOCK_SIZE)
	{
		rc2_encrypt( (RC2_Schedule *)m_extKey, &(inBf[i]), &(outBf[i]) );
	}

	return sae_ok;
}

int cAlgRC2::decryptECB(
		const BYTE *inBf,	// Input encrypted buffer.
		BYTE *outBf,		// Output plaintext buffer.
		int  bfSize			// In/out buffers' size.
		)
{
	if ((bfSize % BLOCK_SIZE) != 0) return sae_invalidBlockSize;

	for( int i = 0; i < bfSize; i += BLOCK_SIZE)
	{
		rc2_decrypt( (RC2_Schedule *)m_extKey, &(outBf[i]), &(inBf[i]) );
	}

	return sae_ok;
}

bool cAlgRC2::test()
{
   //Key length (bytes) = 8
   //Effective key length (bits) = 63
   //Key = 00000000 00000000
   //Plaintext = 00000000 00000000
   //Ciphertext = ebb773f9 93278eff
	
	
	cAlgRC2 rc2( 64 );

	int bits = 63;

	BYTE key[8] = {0};
	BYTE pt[8] = {0};
	BYTE ct[8] = { 0xeb, 0xb7, 0x73, 0xf9, 0x93, 0x27, 0x8e, 0xff };
	BYTE res[8] = {0};

	rc2.expandKey( key, sizeof( key ), bits );
	rc2.encryptECB( pt, res, sizeof( pt ) );

	bool ok = memcmp( res, ct, sizeof( pt ) ) == 0;

	if ( !ok )
		return false;

	rc2.decryptECB( ct, res, sizeof( ct ) );
	ok = memcmp( res, pt, sizeof( pt ) ) == 0;

	if ( !ok )
		return false;

	//Key length (bytes) = 8
	//Effective key length (bits) = 64
	//Key = ffffffff ffffffff
	//Plaintext = ffffffff ffffffff
	//Ciphertext = 278b27e4 2e2f0d49

	bits = 64;
	memset( key, 0xFF, sizeof(key) );
	memset( pt, 0xFF, sizeof(pt) );
	BYTE ct1[] = { 0x27, 0x8b, 0x27, 0xe4, 0x2e, 0x2f, 0x0d, 0x49 };

	rc2.expandKey( key, sizeof( key ), bits );
	rc2.encryptECB( pt, res, sizeof( pt ) );

	ok = memcmp( res, ct1, sizeof( pt ) ) == 0;

	if ( !ok )
		return false;

	rc2.decryptECB( ct1, res, sizeof( ct1 ) );
	ok = memcmp( res, pt, sizeof( pt ) ) == 0;

	if ( !ok )
		return false;

	return ok;
}