/////////////////////////////////////////////////////////////////
//
//      Copyright (c) 2015 Jetico, Inc., Finland
//      All rights reserved.
//
//      File: cAlgRC2.h
//      Revision:
//      Created:
//      Description:
//
/////////////////////////////////////////////////////////////////

#ifndef __cAlgRC2__
#define __cAlgRC2__

#include "ALG_DEF.H"
#include "cSymAlg.h"

class cAlgRC2: public cSymAlg
{
 private:
	BYTE *m_extKey;
	int m_keySize;

	BOOL getAlgConstants( int *blockSize, int *keySize );

 public:

     cAlgRC2( int keyBitLength );
	 ~cAlgRC2();

	static int getBlockSize();

	static cSymAlg *createAlg( BYTE *key, int keySize );  // keySize in bytes
	bool expandKey(
			BYTE *key,
			int keySizeBytes, // key size in bytes
			int effectiveKeyLengthBits = 1024// effective key length in bits
		);


	// Encryption and decription in Electronic Code Book mode

	virtual int encryptECB(
		const BYTE *inBf,		// Input plaintext buffer. 
		BYTE *outBf,	// Output encrypted buffer
		int  bfSize		// In/out buffers' size.
		);

	virtual int decryptECB(
		const BYTE *inBf,		// Input encrypted buffer.
		BYTE *outBf,	// Output plaintext buffer.
		int  bfSize		// In/out buffers' size.
		);

	static bool test();
};

class cAlgRC2_40 : public cAlgRC2
{
public:
	cAlgRC2_40()
		: cAlgRC2( 40 )
	{
	}

	static const char *getDisplayName()
	{
		return "40BitRC2";
	}

	static int getKeySize( int *pMax )
	{
		if (pMax) *pMax = 128 / 8;
		return 40 / 8;
	}

	static cSymAlg * createAlg( BYTE *key, int keySize ) // keySize in bytes
	{
		cAlgRC2 *alg = new cAlgRC2_40();
		if ( alg == NULL )
			return NULL;

		if ( !alg->expandKey( key, keySize, keySize * 8 ) )
		{
			delete alg;
			return NULL;
		}

		return alg;
	}
};

class cAlgRC2_128 : public cAlgRC2
{
public:
	cAlgRC2_128()
		: cAlgRC2( 128 )
	{
	}

	static const char *getDisplayName()
	{
		return "128BitRC2";
	}

	static int getKeySize( int *pMax )
	{
		if (pMax) *pMax = 128 / 8;
		return 128 / 8;
	}

	static cSymAlg * createAlg( BYTE *key, int keySize ) // keySize in bytes
	{
		cAlgRC2 *alg = new cAlgRC2_128();
		if ( alg == NULL )
			return NULL;

		if ( !alg->expandKey( key, keySize, keySize * 8 ) )
		{
			delete alg;
			return NULL;
		}

		return alg;
	}
};

#endif // __cAlgRC2__

