//#include <windows.h>
#include "x509_tags.h"
#include "x509_structs.h"
#include "oidFuncs.h"
#include "ASN1_parseAndPrint.h"
#include "../SMemory/SBlocks.h"
#include "pkcs12_funcs.h"
#include "../SMemory/cPassword.h"
#ifndef maxLen
#define maxLen( buff ) ( sizeof( buff ) / sizeof( *buff ) )
#endif

#include "../parser/pfx_pdu.h"

int P12_PFX_PDU_makeP7Data(
		BYTE *out, int &outSize,
		const BYTE *contentInfo, int contentInfoSize, 
		const BYTE *macSalt, int macSaltSize,
		int iterations,
		cPassword *pwd
	);

///////////////////////////////////////////////////
//
//	Parse and print ANS.1 packet.
//

#include <stdio.h>
#include <string.h>

char *byteStr( BYTE *pkt, int len, int max )
{
	if ( len < 0 )	
		len = max;

	if ( len > max )
		len = max;

	static char buff[200];

	if ( len > sizeof( buff ) / 3 )
		len = sizeof( buff ) / 3;

	for( int i = 0; i < len; i++ )
	{
		sprintf( buff + i*3, "%X%X ", ( pkt[i] >> 4 ) & 0xF, pkt[i] & 0xF );
	}

	if ( len == 0 )
		buff[0] = 0;
	else
		buff[ len*3 - 1 ] = 0;

	return buff;
}

static char *byteSym( char *pos, BYTE b )
{
	sprintf( pos, " %02X", 0xFF & b );
	return pos;
}

static char sizePattern[ 16 ] = { 0 };
static int maxSizeLen = 0;

static char *makeEndOfString( char *buff, x509_struct *s, int recCount )
{
	int len = 14 + maxSizeLen;
	memset( buff, ' ', len );
	buff[ len ] = '\0';

	while( recCount-- )
		strcat( buff, " ." );

	strcat( buff, " End of " );

	if ( s->T & asn1tag_OPTIONAL )
		sprintf( buff + strlen(buff), "[%d]", s->getOptionalNum() );
	else
		strcat( buff, getTagString( s->T ) );

	return buff;
}

static char *makeHeadString( char *buff, x509_struct *s, int recCount, bool *pBodyPrinted )
{
	*pBodyPrinted = false;

	int hdrSize = s->TL - s->L;
	BYTE *hdr = s->B - hdrSize;

	int i;
	for ( i = 0; i <  hdrSize; i++ )
		byteSym( buff + i*3, hdr[i] );

	for ( ; i <  6; i++ )
		strcpy( buff + i*3, "   " );

	if ( sizePattern[0] == 0 ) // first time largers size is to printed.
	{
		char test[ 32 ];
		sprintf( test, "%d", s->L );
		maxSizeLen = strlen( test ) + 1;
		
		sprintf( sizePattern, "%%%dd", maxSizeLen );
	}

	sprintf( buff + 14, sizePattern, s->L );
	//sprintf( buff + 14, "%8d:", s->L );

	while( recCount-- )
		strcat( buff, " ." );

	strcat( buff, " " );

	if ( s->T & asn1tag_OPTIONAL )
		sprintf( buff + strlen(buff), "[%d]", s->getOptionalNum() );
	else
		strcat( buff, getTagString( s->T ) );

//	if ( s->uL )
//		sprintf( buff + strlen( buff ), " ( parsed length - %d )", s->L );
//	else
	if ( !( s->T & asn1tag_CONSTRUCTED ) )
	{
		if ( ( (( s->T & asn1tag_TYPE ) >= asn1tag_BOOLEAN ) && 
			   (( s->T & asn1tag_TYPE ) <= asn1tag_OCTET_STRING ) &&
			   ( s->L <= 1 )
			 ) ||
			 ( (( s->T & asn1tag_TYPE ) == asn1tag_INTEGER ) && 
			   ( s->L <= sizeof(int) )
			 )
		   )
		{
			int l = 0;
			for ( UINT i = 0; i < s->L; i++ ) // big-endian buffer
				l = ( l << 8 ) + s->B[i];

			sprintf( buff + strlen(buff), " %d ( %X )", l, l );
			*pBodyPrinted = true;
		}
		else
		if ((( s->T & asn1tag_TYPE ) >= asn1tag_NumericString ) && 
			(( s->T & asn1tag_TYPE ) <= asn1tag_GeneralizedTime ) &&
			 ( s->L < 100 ) 
			 )
		{
			strcat( buff, "->" );
			int len = strlen(buff);
			strncpy( buff + len, ( const char *)s->B, s->L );
			buff[ len + s->L ] = 0;
			*pBodyPrinted = true;
		}
		else
		if ((( s->T & asn1tag_TYPE ) == asn1tag_UnicodeString ) && 
			 ( s->L < 100 ) 
			 )
		{
			strcat( buff, " - " );
			char *pos = buff + strlen(buff);
			UINT i;
			for ( i = 0; 2 * i < s->L; i++ )
				pos[ i ] = s->B[ 2*i + 1 ];
			pos[i] = 0;

			*pBodyPrinted = true;
		}
		else
		if ( ( s->T & asn1tag_TYPE ) == asn1tag_END_OF_CONTENTS_MARKER )
		{
			if ( ( s->T & asn1tag_OPTIONAL ) == 0 )
				*pBodyPrinted = true;
		}
		else
		if ( ( s->T & asn1tag_TYPE ) == asn1tag_OBJECT_IDENTIFIER )
		{
			char oidStr[ maxOIDStrSize ];
			oid2string( oidStr, s->B, s->L );

			char name[ 512 ];
			bool ok = generateOIDName( name, oidStr ) != 0;

			sprintf( buff + strlen(buff), " { %s }", ok ? name :  oidStr );
			*pBodyPrinted = true;

		}
		else
		if ( s->L == 0 )
			*pBodyPrinted = true;

	}

	return buff;
}

char buff[ 1024 ];

static cPlainTextBlock ptbConstructedOctet;

const int maxOctets = 100;

cPTB singleOctetT[maxOctets];
int singleOctetPosT[maxOctets];
int singleOctetQ = 0;
int singleOctetGetNext( int pos )
{
	if ( singleOctetQ == maxLen( singleOctetT ) )
	{
		printf( "Octet count exceeded!\n" );
		return singleOctetQ - 1;
	}
	
	singleOctetPosT[singleOctetQ] = pos;
	
	return singleOctetQ++;
}

static bool encodeConstructedOctet = false;
static BYTE *start = 0;
//static bool printEOF = false;

static bool printOctetHere = true;

int stupedParser(
	BYTE *ptr,		// Packet begin
	int size,		// size of the packet
	BYTE **pPtr = NULL,	// pointer to next packet
	int *pSize = NULL,	// size left
	int recCount = 0
	)
{
	x509_struct st;
	int res = 0;

	if ( ( size > 0 ) && ( ptr[0] == 0 ) )
	{
		printf( "Zero byte skipped!\n" );
		size--;
		ptr++;
	}

	if ( start == 0 ) start = ptr;

	while ( size != 0 )
	{
		res = st.parse( ptr, size );

		if ( res != x509e_ok )
		{
			printf( "Error %d\n", res );
			return res;
		}

		bool bodyPrinted = false;

		printf( "%04X %s\n", ptr - start, makeHeadString( buff, &st, recCount, &bodyPrinted ) );

		bool bValidBody = true;

		if ( !( st.T & asn1tag_CONSTRUCTED ) && !bodyPrinted )
		{
			x509_struct stl;
			int res = stl.parse( st.B, st.L );
			//bValidBody = ( res == x509e_ok );

			printf( " Octet[%d] -> %s\n", singleOctetQ, byteStr( st.B, st.L, 20) );
			//if ( bValidBody )
			//	printf( "\n", res );
			//else
			//	printf( " Err %d!\n", res );
		}

		if ( st.T & asn1tag_CONSTRUCTED )
		{
			if ( ( st.T &  asn1tag_TYPE ) == asn1tag_OCTET_STRING )
				encodeConstructedOctet = true;

			res = stupedParser( st.B, st.L, NULL, NULL, recCount + 1 );

			encodeConstructedOctet = false;

			if ( res != x509e_ok ) return res;

			//if ( printEOF )
			//{
			//	printf( "%04X                   :", ptr + st.TL - start );
			//	for( int i = recCount; i--; )
			//		printf( " ." );

			//	printf( " EOF %s -> %04X - %d\n", getTagString( st.T ), st.B - ( st.TL - st.L ) - start, st.L );
			//}

		}
		else
		if ( encodeConstructedOctet )
		{
			ptbConstructedOctet.WriteBytes( st.B, st.L );
		}
		else
		if ( ( st.T &  asn1tag_TYPE ) == asn1tag_OCTET_STRING ) 
		{
			if (
				( st.B[0] == 0x30 ) && 
				(( st.L > 8 ) || ( st.B[2] == asn1tag_OBJECT_IDENTIFIER ))  && 
				printOctetHere )
			{
				printf( "--------------PARSED OCTET %04X------------\n", (int)(st.B - start) );
				BYTE *save = start;
				start = 0;

				stupedParser( st.B, st.L );

				start = save;
				printf( "--------------END OF PARSED OCTET %04X------------\n", (int)(st.B - start) );
			}
			else
			if ( bValidBody )
				singleOctetT[singleOctetGetNext(ptr - start)].WriteBytes( st.B, st.L );

		}
		else
		if ( ( st.T &  asn1tag_TYPE ) == asn1tag_BIT_STRING ) 
		{
			if (
				( st.B[1] == 0x30 ) && 
				(( st.L > 9 ) || ( st.B[3] == asn1tag_OBJECT_IDENTIFIER ))  && 
				printOctetHere )
			{
				printf( "--------------PARSED BIT STRING %04X------------\n", (int)(st.B - start) );
				BYTE *save = start;
				start = 0;

				stupedParser( st.B + 1, st.L - 1 );

				start = save;
				printf( "--------------END OF PARSED BIT STRING %04X------------\n", (int)(st.B - start) );
			}
			else
			if ( bValidBody )
				singleOctetT[singleOctetGetNext(ptr - start)].WriteBytes( st.B, st.L );
		}
		else
		if ( ( st.T == asn1tag_OPTIONAL ) && bValidBody )
			singleOctetT[singleOctetGetNext(ptr - start)].WriteBytes( st.B, st.L );

		ptr += st.TL;
		size -= st.TL;

		if ( size == 0 )
		{
			if ( st.T & asn1tag_CONSTRUCTED )
				printf( "%04X %s %04X\n", ptr - start, makeEndOfString( buff, &st, recCount ), st.H - start );

			break;
		}

		if ( size < 0 )
			return x509e_invalidPktSize;

		bValidBody = true;
	}

	if ( pPtr  ) *pPtr = ptr;
	if ( pSize ) *pSize = size;

	return x509e_ok;
}

#ifdef _DEBUG
#ifdef _WIN32
static int writeToFile( HANDLE hFile , BYTE *ptr, DWORD size, bool append )
{
	if ( hFile == INVALID_HANDLE_VALUE )
	{
		return 0;
	}

	DWORD pos;

	if ( append )
		pos = SetFilePointer( hFile, 0, NULL, FILE_END );

	if ( !WriteFile( hFile, ptr, size, &size, NULL ) )
		size = 0;

	CloseHandle( hFile );

	return size;
}

static int WriteToFile( const char *fileName, BYTE *ptr, DWORD size, bool append=false )
{
	HANDLE hFile = CreateFileA(
				fileName,
				GENERIC_WRITE | ( append ? GENERIC_READ : 0 ),
				FILE_SHARE_READ,
				NULL,
				append ? OPEN_ALWAYS : CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL
			);

	return writeToFile( hFile, ptr, size, append );
}

static int parseOctet( int i )
{
	cPTB &ptb = singleOctetT[i];

	x509_struct stl;
	int res = stl.parse( ptb.buffer, ptb.bufferSize );
	if ( res != x509e_ok )
		return res;

	start = 0;
	printf( "\nOctet[%d] %04X, size %d\n", i, singleOctetPosT[i], ptb.bufferSize );
	printf( " -> %s\n\n", byteStr( ptb.buffer, ptb.bufferSize, 24) );

	res = stupedParser( ptb.buffer, ptb.bufferSize );

	return res;
}

int ASN1_parseAndPrintUnit( BYTE *ptr, int size, bool skipOctetParsing, const char *szPassword )
{
	printOctetHere = !skipOctetParsing;
	
	int lSize = size;

	int res = stupedParser( ptr, size, NULL, &lSize );
	if ( res != x509e_ok ) return res;

	if ( ptbConstructedOctet.bufferSize  == 0 )
	{
		int i = 0;
		for( ; i < singleOctetQ; i++ )
		{
			res = parseOctet(i);
		}

		if ( szPassword == NULL )
			return res;

		cPassword pwd( szPassword );

		//if ( res != ppe_ok ) return res;

		cPTB ptbT[2];

		for( int number = 0;; number++ )
		{
			res = P12_PFX_PDU_parse( ptr, size, &ptbT[number], &pwd, NULL, number );

			if ( res == x509e_emptyParameters )
			{
				printf( "\nPFX_PDU total %d\n", number );
				res = ppe_ok;
				break;
			}
			
			if ( res != ppe_ok )
				break;

			printf( "\nPFX_PDU %d, size %d:\n\n", number, ptbT[number].bufferSize );
			start = 0;
			res = stupedParser( ptbT[number].buffer, ptbT[number].bufferSize );
			
			for( ; i < singleOctetQ; i++ )
			{
				// start = 0;
				res = parseOctet( i );
			}
		}

		// Step 4. Prepare encrypted certificate packet as p7encryptedData
		//
		//	Problem was found -> Certificate bag attribute was added to entire certificate
		//
		//const BYTE *encryptedCrt = singleOctetT[4].buffer;
		//int encryptedCrtSize = singleOctetT[4].bufferSize;

		cPTB &ptbCrt = singleOctetT[9];
		
		cPTB ptbCrtBag;

		res = P12_makeCertificateKeyBagPkt( ptbCrtBag, ptbCrt.buffer, ptbCrt.bufferSize );
		if ( res != pkae_ok ) return res;

		if ( !ptbCrtBag.equ( ptbT[0] ) )
		{
			start = 0;
			stupedParser( ptbCrtBag.buffer, ptbCrtBag.bufferSize );
			return x509e_internalError;
		}


		int iterations = 2048;
		const char *pbeAlgOID = oidP12PbeWithSHAAnd40BitRC2_CBC;
		cPTB &origSalt = singleOctetT[3];

		//cPTB ptbCertBagEnc;
		//res = P12_makePKCS7EncryptedDataPkt( ptbCertBagEnc, ptbCrtBag.buffer, ptbCrtBag.bufferSize, origSalt.buffer, origSalt.bufferSize, &pwd, iterations, pbeAlgOID );
		//if ( res != x509e_ok )	return res;

		cPTB encData;
		res = P12_makePBEncryptedBuffer( encData, ptbCrtBag.buffer, ptbCrtBag.bufferSize, origSalt.buffer, origSalt.bufferSize, &pwd, iterations, pbeAlgOID );
		if ( res != pkae_ok ) return res;

		if ( !encData.equ( singleOctetT[4] ) )
			return x509e_internalError;

		const BYTE *encryptedCrt = encData.buffer;
		int encryptedCrtSize = encData.bufferSize;

		// Step 3. Prepare encrypted secret key as PKCS8ShroudedKeyBag
		//	Shrouded key bag is now generated properly, the local key id generationwas changed only.
		//cPTB &origPKCS8ShroudedKeyBag = singleOctetT[5];
		cPTB ptbPKCS8ShroudedKeyBag;

		//cPTB &origKey = ptbT[1]; // Second parsed PFX_PDU.

		res = makePKCS8ShroudedKeyBagPkt( ptbPKCS8ShroudedKeyBag, singleOctetT[5], ptbCrt, singleOctetT, singleOctetQ, &pwd );
		if ( res != x509e_ok )
			return res;

		int p7dataContentInfoSize = ptbPKCS8ShroudedKeyBag.bufferSize;
		const BYTE *p7dataContentInfo = ptbPKCS8ShroudedKeyBag.buffer;


		// Step 2. pfx_pdu content info is generated properly.
		//	From encrypted certificate and p7data content info with encrypted secret key

		BYTE macSalt[8] = { 0 };
		memcpy( macSalt, singleOctetT[2].buffer, sizeof( macSalt ) );	

		// 1. p7encryptedData
		//const BYTE *encryptedCrt = singleOctetT[4].buffer;
		//int encryptedCrtSize = singleOctetT[4].bufferSize;

		// 2. p7data
		//const BYTE *p7dataContentInfo = singleOctetT[5].buffer;
		//int p7dataContentInfoSize = singleOctetT[5].bufferSize;

		int contentInfoSize = 0;
		res = makeContentInfo( NULL, contentInfoSize, encryptedCrt, encryptedCrtSize, origSalt.buffer, origSalt.bufferSize, &pwd, iterations, pbeAlgOID, p7dataContentInfo, p7dataContentInfoSize );
		if ( res != x509e_bufferTooSmall )
			return res;

		cPTB pduCI;
		if ( pduCI.Allocate( contentInfoSize ) == NULL )
			return x509e_memoryAllocation;

		BYTE *contentInfo = pduCI.buffer;

		res = makeContentInfo( contentInfo, contentInfoSize, encryptedCrt, encryptedCrtSize, origSalt.buffer, origSalt.bufferSize, &pwd, iterations, pbeAlgOID, p7dataContentInfo, p7dataContentInfoSize );
		if ( res != x509e_ok )
			return res;

		if ( !pduCI.equ( singleOctetT[0] ) )
		{
			start = 0;
			stupedParser( pduCI.buffer, pduCI.bufferSize );
			return x509e_invalidPkt;
		}

		// Step 1. pfx_pdu is generated properly.

		//BYTE *contentInfo = singleOctetT[0].buffer;
		//int contentInfoSize = singleOctetT[0].bufferSize;

		int pfx_pdu_size = 0;
		res = P12_PFX_PDU_makeP7Data( NULL, pfx_pdu_size, contentInfo, contentInfoSize, macSalt, sizeof( macSalt ), 2048, &pwd );
		if ( res != x509e_bufferTooSmall )
			return res;

		cPTB pdu;
		if ( pdu.Allocate( pfx_pdu_size ) == NULL )
			return x509e_memoryAllocation;

		res = P12_PFX_PDU_makeP7Data( pdu.buffer, pfx_pdu_size, contentInfo, contentInfoSize, macSalt, sizeof( macSalt ), 2048, &pwd );
		if ( res != x509e_ok )
			return res;

		if ( ( pfx_pdu_size != size ) || 
			 ( memcmp( pdu.buffer, ptr, pfx_pdu_size ) != 0 ) )
		{
			return x509e_invalidPkt;
		}

		if ( WriteToFile( "d:\\tmp\\gen.p12", pdu.buffer, pfx_pdu_size ) != pfx_pdu_size )
			return GetLastError();

		return res;
	}

	cPlainTextBlock ptbCopy;

	ptbCopy.Assign( ptbConstructedOctet.buffer, ptbConstructedOctet.bufferSize );
	ptbConstructedOctet.SeekToBegin();
	start = 0;

	printf( "\nParses Constructed Octet 1\n" );

	res = stupedParser( ptbCopy.buffer, ptbCopy.bufferSize );
	if ( res != x509e_ok ) return res;

	ptbCopy.Assign( ptbConstructedOctet.buffer, ptbConstructedOctet.bufferSize );
	ptbConstructedOctet.SeekToBegin();
	start = 0;

	printf( "\nParses Constructed Octet 2\n" );

	res = stupedParser( ptbCopy.buffer, ptbCopy.bufferSize );
	WriteToFile( "Constructed Octet 2.bin", ptbCopy.buffer, ptbCopy.bufferSize );

	return res;
}
#endif // _WIN32
#endif // _DEBUG

//
//	End of 'parse and print' ANS.1 packet.
//
///////////////////////////////////////////////////
